# -*- coding: utf-8 -*-
###########################################################################
#                                                                         #
#   Based on work by Alain "Chiaroscuro" Ducharme,  Campbell Barton,      #
#    Jiri Hnidek, Paolo Ciccone, Enrico Cerica and Marc Ume               #
#                                                                         #
#   Made by David "Livuxman" Rodrguez                                    #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program.  If not, see <http://www.gnu.org/licenses/>. #
#                                                                         #
#                                                                         #
###########################################################################
#
# Octane render engine Blender 2.5 plug-in
# v0.1


import bpy
import mathutils
import os
import sys
import time

#export obj part


def matrix_vect(mat,vec):
        vecr = [0.,0.,0.,0.,0.]
        for i in range(4):
                for j in range(4):
                        vecr[i] += vec[j] * mat[j][i] 
        return vecr
                
def rotate90x(vect):
        return [vect[0],vect[2],-vect[1]]
        
def splitExt(path):
    dotidx = path.rfind('.')
    if dotidx == -1:
        return path, ''
    else:
        return path[:dotidx], path[dotidx:]

def fixName(name):
    if name == None:
        return 'None'
    else:
        return name.replace(' ', '_')

def write_mtl(scene, filename, copy_images, mtl_dict):

    world = scene.world
    worldAmb = world.ambient_color

    dest_dir = os.path.dirname(filename)

    def copy_image(image):
        fn = bpy.utils.expandpath(image.filename)
        fn_strip = os.path.basename(fn)
        if copy_images:
            rel = fn_strip
            fn_abs_dest = os.path.join(dest_dir, fn_strip)
            if not os.path.exists(fn_abs_dest):
                shutil.copy(fn, fn_abs_dest)
        else:
            rel = fn

        return rel


    file = open(filename, "w")
    # XXX
#       file.write('# Blender MTL File: %s\n' % Blender.Get('filename').split('\\')[-1].split('/')[-1])
    file.write('# Material Count: %i\n' % len(mtl_dict))
    # Write material/image combinations we have used.
    for key, (mtl_mat_name, mat, img) in mtl_dict.items():

        # Get the Blender data for the material and the image.
        # Having an image named None will make a bug, dont do it :)

        file.write('newmtl %s\n' % mtl_mat_name) # Define a new material: matname_imgname

        if mat:
              #file.write('Ns %.6f\n' % ((mat.getHardness()-1) * 1.9607843137254901) ) # Hardness, convert blenders 1-511 to MTL's
              file.write('Ns %.6f\n' % (mat.specular_hardness*10) ) # Hardness, convert blenders 1-511 to MTL's
              file.write('Ka %.6f %.6f %.6f\n' %  tuple([c*mat.ambient for c in worldAmb])  ) # Ambient, uses mirror colour,
              #file.write('Kd %.6f %.6f %.6f\n' % tuple([c*mat.ref for c in mat.rgbCol]) ) # Diffuse
              file.write('Kd %.6f %.6f %.6f\n' % tuple([c for c in mat.diffuse_color]) ) # Diffuse
              file.write('Ks %.6f %.6f %.6f\n' % tuple([(c) for c in mat.specular_color]) ) # Specular
              file.write('Ni %.6f\n' % mat.raytrace_transparency.ior) # Refraction index
              #file.write('Ks %.6f %.6f %.6f\n' % tuple([c*mat.spec for c in mat.specCol]) ) # Specular
              
              file.write('d %.6f\n' % mat.alpha) # Alpha (obj uses 'd' for dissolve)
              # 0 to disable lighting, 1 for ambient & diffuse only (specular color set to black), 2 for full lighting.
              if mat.shadeless:
                      file.write('illum 0\n') # ignore lighting
              elif mat.specular_intensity == 0:
                      file.write('illum 1\n') # no specular.
              else:
                      file.write('illum 2\n') # light normaly

        else:
            #write a dummy material here?
            file.write('Ns 0\n')
            file.write('Ka %.6f %.6f %.6f\n' %  tuple([c for c in worldAmb])  ) # Ambient, uses mirror colour,
            file.write('Kd 0.8 0.8 0.8\n')
            file.write('Ks 0.8 0.8 0.8\n')
            file.write('d 1\n') # No alpha
            file.write('illum 2\n') # light normaly

        # Write images!
        if img:  # We have an image on the face!
            # write relative image path
            rel = copy_image(img)
            file.write('map_Kd %s\n' % rel) # Diffuse mapping image

        elif mat: # No face image. if we havea material search for MTex image.
            for mtex in mat.texture_slots:
                if mtex and mtex.texture.type == 'IMAGE':
                    try:
                        filename = copy_image(mtex.texture.image)
                        file.write('map_Kd %s\n' % filename) # Diffuse mapping image
                        break
                    except:
                        # Texture has no image though its an image type, best ignore.
                        pass

        file.write('\n\n')

    file.close()
    
def test_nurbs_compat(ob):
    if ob.type != 'Curve':
        return False

    for nu in ob.data:
        if (not nu.knotsV) and nu.type != 1: # not a surface and not bezier
            return True

    return False


# XXX not converted
def write_nurb(file, ob, ob_mat):
    tot_verts = 0
    cu = ob.data

    # use negative indices
    Vector = Blender.mathutils.Vector
    for nu in cu:

        if nu.type==0:          DEG_ORDER_U = 1
        else:                           DEG_ORDER_U = nu.orderU-1  # Tested to be correct

        if nu.type==1:
            print("\tWarning, bezier curve:", ob.name, "only poly and nurbs curves supported")
            continue

        if nu.knotsV:
            print("\tWarning, surface:", ob.name, "only poly and nurbs curves supported")
            continue

        if len(nu) <= DEG_ORDER_U:
            print("\tWarning, orderU is lower then vert count, skipping:", ob.name)
            continue

        pt_num = 0
        do_closed = (nu.flagU & 1)
        do_endpoints = (do_closed==0) and (nu.flagU & 2)

        for pt in nu:
            pt = Vector(pt[0], pt[1], pt[2]) * ob_mat
            file.write('v %.6f %.6f %.6f\n' % (pt[0], pt[1], pt[2]))
            pt_num += 1
        tot_verts += pt_num

        file.write('g %s\n' % (fixName(ob.name))) # fixName(ob.getData(1)) could use the data name too
        file.write('cstype bspline\n') # not ideal, hard coded
        file.write('deg %d\n' % DEG_ORDER_U) # not used for curves but most files have it still

        curve_ls = [-(i+1) for i in range(pt_num)]

        # 'curv' keyword
        if do_closed:
            if DEG_ORDER_U == 1:
                pt_num += 1
                curve_ls.append(-1)
            else:
                pt_num += DEG_ORDER_U
                curve_ls = curve_ls + curve_ls[0:DEG_ORDER_U]

        file.write('curv 0.0 1.0 %s\n' % (' '.join( [str(i) for i in curve_ls] ))) # Blender has no U and V values for the curve

        # 'parm' keyword
        tot_parm = (DEG_ORDER_U + 1) + pt_num
        tot_parm_div = float(tot_parm-1)
        parm_ls = [(i/tot_parm_div) for i in range(tot_parm)]

        if do_endpoints: # end points, force param
            for i in range(DEG_ORDER_U+1):
                parm_ls[i] = 0.0
                parm_ls[-(1+i)] = 1.0

        file.write('parm u %s\n' % ' '.join( [str(i) for i in parm_ls] ))

        file.write('end\n')

    return tot_verts

def write(filename, objects, scene,
          EXPORT_TRI=False,
          EXPORT_EDGES=False,
          EXPORT_NORMALS=False,
          EXPORT_NORMALS_HQ=False,
          EXPORT_UV=True,
          EXPORT_MTL=True,
          EXPORT_COPY_IMAGES=False,
          EXPORT_APPLY_MODIFIERS=True,
          EXPORT_ROTX90=True,
          EXPORT_BLEN_OBS=True,
          EXPORT_GROUP_BY_OB=False,
          EXPORT_GROUP_BY_MAT=False,
          EXPORT_KEEP_VERT_ORDER=False,
          EXPORT_POLYGROUPS=False,
          EXPORT_CURVE_AS_NURBS=True):
    '''
    Basic write function. The context and options must be alredy set
    This can be accessed externaly
    eg.
    write( 'c:\\test\\foobar.obj', Blender.Object.GetSelected() ) # Using default options.
    '''

    # XXX
    import math

    unitFactor = 1

    def veckey3d(v):
        return round(v.x, 6), round(v.y, 6), round(v.z, 6)

    def veckey2d(v):
        return round(v[0], 6), round(v[1], 6)
        # return round(v.x, 6), round(v.y, 6)

    def findVertexGroupName(face, vWeightMap):
        """
        Searches the vertexDict to see what groups is assigned to a given face.
        We use a frequency system in order to sort out the name because a given vetex can
        belong to two or more groups at the same time. To find the right name for the face
        we list all the possible vertex group names with their frequency and then sort by
        frequency in descend order. The top element is the one shared by the highest number
        of vertices is the face's group
        """
        weightDict = {}
        for vert_index in face.verts:
#               for vert in face:
            vWeights = vWeightMap[vert_index]
#                       vWeights = vWeightMap[vert]
            for vGroupName, weight in vWeights:
                weightDict[vGroupName] = weightDict.get(vGroupName, 0) + weight

        if weightDict:
            alist = [(weight,vGroupName) for vGroupName, weight in weightDict.items()] # sort least to greatest amount of weight
            alist.sort()
            return(alist[-1][1]) # highest value last
        else:
            return '(null)'

    # TODO: implement this in C? dunno how it should be called...
    def getVertsFromGroup(me, group_index):
        ret = []

        for i, v in enumerate(me.verts):
            for g in v.groups:
                if g.group == group_index:
                    ret.append((i, g.weight))

        return ret


    print('OBJ Export path: "%s"' % filename)
    temp_mesh_name = '~tmp-mesh'
    time1 = time.clock()
#    time1 = sys.time()
#    scn = Scene.GetCurrent()

    file = open(filename, "w")
    # Write Header
    file.write('# Blender v%s OBJ File: %s\n' % (bpy.app.version_string, bpy.data.filename.split('/')[-1].split('\\')[-1] ))
    file.write('# www.blender.org\n')

    # Tell the obj file what material file to use.
    if EXPORT_MTL:
        mtlfilename = '%s.mtl' % '.'.join(filename.split('.')[:-1])
        file.write('mtllib %s\n' % ( mtlfilename.split('\\')[-1].split('/')[-1] ))

    if EXPORT_ROTX90:
        mat_xrot90= mathutils.RotationMatrix(-math.pi/2, 4, 'X')

    # Initialize totals, these are updated each object
    totverts = totuvco = totno = 1

    face_vert_index = 1

    globalNormals = {}

    # A Dict of Materials
    # (material.name, image.name):matname_imagename # matname_imagename has gaps removed.
    mtl_dict = {}

    # Get all meshes
    for ob_main in objects:
#        if ob_main.parent and ob_main.parent.duplis_used:
#          print(ob_main.type, 'duplis ')
        #,ob_main.parent.dupli_type)
        # ignore dupli children
        if ob_main.parent and ob_main.parent.dupli_type != 'NONE':
            # XXX
            print(ob_main.name, 'is a dupli child - ignoring')
            continue

        obs = []
        if ob_main.dupli_type != 'NONE':
            # XXX
            print('creating dupli_list on', ob_main.name)
            ob_main.create_dupli_list(scene)

            obs = [(dob.object, dob.matrix) for dob in ob_main.dupli_list]

            # XXX debug print
            print(ob_main.name, 'has', len(obs), 'dupli children')
        else:
            obs = [(ob_main, ob_main.matrix)]

        for ob, ob_mat in obs:

            # XXX postponed
#                       # Nurbs curve support
#                       if EXPORT_CURVE_AS_NURBS and test_nurbs_compat(ob):
#                               if EXPORT_ROTX90:
#                                       ob_mat = ob_mat * mat_xrot90

#                               totverts += write_nurb(file, ob, ob_mat)

#                               continue
#                       end nurbs

            if ob.type != 'MESH':
                continue

            me = ob.create_mesh(scene, EXPORT_APPLY_MODIFIERS, 'PREVIEW')
#                       # Will work for non meshes now! :)
#            me= BPyMesh.getMeshFromObject(ob, none, True, True, scene)
            if not me:
                continue

            if EXPORT_ROTX90:
                me.transform(mat_xrot90 * ob_mat)
            else:
                me.transform(ob_mat)


            if EXPORT_UV:
                faceuv = len(me.uv_textures) > 0
            else:
                faceuv = False

            # XXX - todo, find a better way to do triangulation
            # ...removed convert_to_triface because it relies on editmesh
            '''
            # We have a valid mesh
            if EXPORT_TRI and me.faces:
                # Add a dummy object to it.
                has_quads = False
                for f in me.faces:
                    if f.verts[3] != 0:
                        has_quads = True
                        break

                if has_quads:
                    newob = bpy.data.objects.new('temp_object', me)
                    # if we forget to set Object.data - crash
                    scene.objects.link(newob)
                    newob.convert_to_triface(scene)
                    # mesh will still be there
                    scene.objects.unlink(newob)
            '''

            # Make our own list so it can be sorted to reduce context switching
            face_index_pairs = [ (face, index) for index, face in enumerate(me.faces)]
            # faces = [ f for f in me.faces ]

            if EXPORT_EDGES:
                edges = me.edges
            else:
                edges = []

            if not (len(face_index_pairs)+len(edges)+len(me.verts)): # Make sure there is somthing to write

                # clean up
                bpy.data.meshes.remove(me)

                continue # dont bother with this mesh.

            # XXX
            # High Quality Normals
            if EXPORT_NORMALS and face_index_pairs:
                me.calc_normals()
#                               if EXPORT_NORMALS_HQ:
#                                       BPyMesh.meshCalcNormals(me)
#                               else:
#                                       # transforming normals is incorrect
#                                       # when the matrix is scaled,
#                                       # better to recalculate them
#                                       me.calcNormals()

            materials = me.materials

            materialNames = []
            materialItems = [m for m in materials]
            if materials:
                for mat in materials:
                    if mat: # !=None
                        materialNames.append(mat.name)
                    else:
                        materialNames.append(None)
                # Cant use LC because some materials are None.
                # materialNames = map(lambda mat: mat.name, materials) # Bug Blender, dosent account for null materials, still broken.

            # Possible there null materials, will mess up indicies
            # but at least it will export, wait until Blender gets fixed.
            materialNames.extend((16-len(materialNames)) * [None])
            materialItems.extend((16-len(materialItems)) * [None])

            # Sort by Material, then images
            # so we dont over context switch in the obj file.
            if EXPORT_KEEP_VERT_ORDER:
                pass
            elif faceuv:
                # XXX update
                tface = me.active_uv_texture.data

                face_index_pairs.sort(key=lambda a: (a[0].material_index, hash(tface[a[1]].image), a[0].smooth))
            elif len(materials) > 1:
                face_index_pairs.sort(key = lambda a: (a[0].material_index, a[0].smooth))
            else:
                # no materials
                face_index_pairs.sort(key = lambda a: a[0].smooth)
#                       if EXPORT_KEEP_VERT_ORDER:
#                               pass
#                       elif faceuv:
#                               try:    faces.sort(key = lambda a: (a.mat, a.image, a.smooth))
#                               except: faces.sort(lambda a,b: cmp((a.mat, a.image, a.smooth), (b.mat, b.image, b.smooth)))
#                       elif len(materials) > 1:
#                               try:    faces.sort(key = lambda a: (a.mat, a.smooth))
#                               except: faces.sort(lambda a,b: cmp((a.mat, a.smooth), (b.mat, b.smooth)))
#                       else:
#                               # no materials
#                               try:    faces.sort(key = lambda a: a.smooth)
#                               except: faces.sort(lambda a,b: cmp(a.smooth, b.smooth))

            faces = [pair[0] for pair in face_index_pairs]

            # Set the default mat to no material and no image.
            contextMat = (0, 0) # Can never be this, so we will label a new material teh first chance we get.
            contextSmooth = None # Will either be true or false,  set bad to force initialization switch.

            if EXPORT_BLEN_OBS or EXPORT_GROUP_BY_OB:
                name1 = ob.name
                name2 = ob.data.name
                if name1 == name2:
                    obnamestring = fixName(name1)
                else:
                    obnamestring = '%s_%s' % (fixName(name1), fixName(name2))

                if EXPORT_BLEN_OBS:
                    file.write('o %s\n' % obnamestring) # Write Object name
                else: # if EXPORT_GROUP_BY_OB:
                    file.write('g %s\n' % obnamestring)


            # Vert
            for v in me.verts:
#                file.write('v %.6f %.6f %.6f\n' % tuple(v.co))
                file.write('v %.6f %.6f %.6f\n' % tuple(v.co*unitFactor))
            # UV
            if faceuv:
                uv_face_mapping = [[0,0,0,0] for f in faces] # a bit of a waste for tri's :/

                uv_dict = {} # could use a set() here
                uv_layer = me.active_uv_texture
                for f, f_index in face_index_pairs:

                    tface = uv_layer.data[f_index]

                    # workaround, since tface.uv iteration is wrong atm
                    uvs = tface.uv
                    # uvs = [tface.uv1, tface.uv2, tface.uv3]

                    # # add another UV if it's a quad
                    # if len(f.verts) == 4:
                    #   uvs.append(tface.uv4)

                    for uv_index, uv in enumerate(uvs):
                        uvkey = veckey2d(uv)
                        try:
                            uv_face_mapping[f_index][uv_index] = uv_dict[uvkey]
                        except:
                            uv_face_mapping[f_index][uv_index] = uv_dict[uvkey] = len(uv_dict)
                            file.write('vt %.6f %.6f\n' % tuple(uv))

#                               uv_dict = {} # could use a set() here
#                               for f_index, f in enumerate(faces):

#                                       for uv_index, uv in enumerate(f.uv):
#                                               uvkey = veckey2d(uv)
#                                               try:
#                                                       uv_face_mapping[f_index][uv_index] = uv_dict[uvkey]
#                                               except:
#                                                       uv_face_mapping[f_index][uv_index] = uv_dict[uvkey] = len(uv_dict)
#                                                       file.write('vt %.6f %.6f\n' % tuple(uv))

                uv_unique_count = len(uv_dict)
#                               del uv, uvkey, uv_dict, f_index, uv_index
                # Only need uv_unique_count and uv_face_mapping

            # NORMAL, Smooth/Non smoothed.
            if EXPORT_NORMALS:
                for f in faces:
                    if f.smooth:
                        for vIdx in f.verts:
                            v = me.verts[vIdx]
                            noKey = veckey3d(v.normal)
                            if noKey not in globalNormals:
                                globalNormals[noKey] = totno
                                totno +=1
                                file.write('vn %.6f %.6f %.6f\n' % noKey)
                    else:
                        # Hard, 1 normal from the face.
                        noKey = veckey3d(f.normal)
                        if noKey not in globalNormals:
                            globalNormals[noKey] = totno
                            totno +=1
                            file.write('vn %.6f %.6f %.6f\n' % noKey)

            if not faceuv:
                f_image = None

            # XXX
            if EXPORT_POLYGROUPS:
                # Retrieve the list of vertex groups
#                               vertGroupNames = me.getVertGroupNames()

                currentVGroup = ''
                # Create a dictionary keyed by face id and listing, for each vertex, the vertex groups it belongs to
                vgroupsMap = [[] for _i in range(len(me.verts))]
#                               vgroupsMap = [[] for _i in xrange(len(me.verts))]
                for g in ob.vertex_groups:
#                               for vertexGroupName in vertGroupNames:
                    for vIdx, vWeight in getVertsFromGroup(me, g.index):
#                                       for vIdx, vWeight in me.getVertsFromGroup(vertexGroupName, 1):
                        vgroupsMap[vIdx].append((g.name, vWeight))

            for f_index, f in enumerate(faces):
                f_v = [{"index": index, "vertex": me.verts[index]} for index in f.verts]

                # if f.verts[3] == 0:
                #       f_v.pop()

#                               f_v= f.v
                f_smooth= f.smooth
                f_mat = min(f.material_index, len(materialNames)-1)
#                               f_mat = min(f.mat, len(materialNames)-1)
                if faceuv:

                    tface = me.active_uv_texture.data[face_index_pairs[f_index][1]]

                    f_image = tface.image
                    f_uv = tface.uv
                    # f_uv= [tface.uv1, tface.uv2, tface.uv3]
                    # if len(f.verts) == 4:
                    #   f_uv.append(tface.uv4)
#                                       f_image = f.image
#                                       f_uv= f.uv

                # MAKE KEY
                if faceuv and f_image: # Object is always true.
                    key = materialNames[f_mat],  f_image.name
                else:
                    key = materialNames[f_mat],  None # No image, use None instead.

                # Write the vertex group
                if EXPORT_POLYGROUPS:
                    if len(ob.vertex_groups):
                        # find what vertext group the face belongs to
                        theVGroup = findVertexGroupName(f,vgroupsMap)
                        if      theVGroup != currentVGroup:
                            currentVGroup = theVGroup
                            file.write('g %s\n' % theVGroup)
#                               # Write the vertex group
#                               if EXPORT_POLYGROUPS:
#                                       if vertGroupNames:
#                                               # find what vertext group the face belongs to
#                                               theVGroup = findVertexGroupName(f,vgroupsMap)
#                                               if      theVGroup != currentVGroup:
#                                                       currentVGroup = theVGroup
#                                                       file.write('g %s\n' % theVGroup)

                # CHECK FOR CONTEXT SWITCH
                if key == contextMat:
                    pass # Context alredy switched, dont do anything
                else:
                    if key[0] == None and key[1] == None:
                        # Write a null material, since we know the context has changed.
                        if EXPORT_GROUP_BY_MAT:
                            # can be mat_image or (null)
                            file.write('g %s_%s\n' % (fixName(ob.name), fixName(ob.data.name)) ) # can be mat_image or (null)
                        file.write('usemtl (null)\n') # mat, image

                    else:
                        mat_data= mtl_dict.get(key)
                        if not mat_data:
                            # First add to global dict so we can export to mtl
                            # Then write mtl

                            # Make a new names from the mat and image name,
                            # converting any spaces to underscores with fixName.

                            # If none image dont bother adding it to the name
#                            if key[1] == None:
                                mat_data = mtl_dict[key] = ('%s'%fixName(key[0])), materialItems[f_mat], f_image
 #                           else:
  #                              mat_data = mtl_dict[key] = ('%s_%s' % (fixName(key[0]), fixName(key[1]))), materialItems[f_mat], f_image

                        if EXPORT_GROUP_BY_MAT:
                            file.write('g %s_%s_%s\n' % (fixName(ob.name), fixName(ob.data.name), mat_data[0]) ) # can be mat_image or (null)

                        file.write('usemtl %s\n' % mat_data[0]) # can be mat_image or (null)

                contextMat = key
                if f_smooth != contextSmooth:
                    if f_smooth: # on now off
                        file.write('s 1\n')
                        contextSmooth = f_smooth
                    else: # was off now on
                        file.write('s off\n')
                        contextSmooth = f_smooth

                file.write('f')
                if faceuv:
                    if EXPORT_NORMALS:
                        if f_smooth: # Smoothed, use vertex normals
                            for vi, v in enumerate(f_v):
                                file.write( ' %d/%d/%d' % \
                                                (v["index"] + totverts,
                                                 totuvco + uv_face_mapping[f_index][vi],
                                                 globalNormals[ veckey3d(v["vertex"].normal) ]) ) # vert, uv, normal

                        else: # No smoothing, face normals
                            no = globalNormals[ veckey3d(f.normal) ]
                            for vi, v in enumerate(f_v):
                                file.write( ' %d/%d/%d' % \
                                                (v["index"] + totverts,
                                                 totuvco + uv_face_mapping[f_index][vi],
                                                 no) ) # vert, uv, normal
                    else: # No Normals
                        for vi, v in enumerate(f_v):
                            file.write( ' %d/%d' % (\
                              v["index"] + totverts,\
                              totuvco + uv_face_mapping[f_index][vi])) # vert, uv

                    face_vert_index += len(f_v)

                else: # No UV's
                    if EXPORT_NORMALS:
                        if f_smooth: # Smoothed, use vertex normals
                            for v in f_v:
                                file.write( ' %d//%d' %
                                            (v["index"] + totverts, globalNormals[ veckey3d(v["vertex"].normal) ]) )
                        else: # No smoothing, face normals
                            no = globalNormals[ veckey3d(f.normal) ]
                            for v in f_v:
                                file.write( ' %d//%d' % (v["index"] + totverts, no) )
                    else: # No Normals
                        for v in f_v:
                            file.write( ' %d' % (v["index"] + totverts) )

                file.write('\n')

            # Write edges.
            if EXPORT_EDGES:
                for ed in edges:
                    if ed.loose:
                        file.write('f %d %d\n' % (ed.verts[0] + totverts, ed.verts[1] + totverts))

            # Make the indicies global rather then per mesh
            totverts += len(me.verts)
            if faceuv:
                totuvco += uv_unique_count

            # clean up
            bpy.data.meshes.remove(me)

        if ob_main.dupli_type != 'NONE':
            ob_main.free_dupli_list()

    file.close()


    # Now we have all our materials, save them
    if EXPORT_MTL:
        write_mtl(scene, mtlfilename, EXPORT_COPY_IMAGES, mtl_dict)
#       if EXPORT_COPY_IMAGES:
#               dest_dir = os.path.basename(filename)
# #             dest_dir = filename
# #             # Remove chars until we are just the path.
# #             while dest_dir and dest_dir[-1] not in '\\/':
# #                     dest_dir = dest_dir[:-1]
#               if dest_dir:
#                       copy_images(dest_dir, mtl_dict)
#               else:
#                       print('\tError: "%s" could not be used as a base for an image path.' % filename)

    print("OBJ Export time: %.2f" % (time.clock() - time1))
#    print "OBJ Export time: %.2f" % (sys.time() - time1)


#end obj part



def oct_properties():
  # Add Octane properties
  # (should be: bpy.types.RenderSettings; doesn't work yet)
  StringProperty = bpy.types.Scene.StringProperty
  FloatProperty = bpy.types.Scene.FloatProperty
  IntProperty = bpy.types.Scene.IntProperty
  BoolProperty = bpy.types.Scene.BoolProperty
  EnumProperty = bpy.types.Scene.EnumProperty
  FloatVectorProperty = bpy.types.Scene.FloatVectorProperty

  StringProperty(attr="oct_path", name="Octane Path",
      description="Full path to Octane's executable",
      default="", maxlen=1024, subtype="FILE_PATH")

  StringProperty(attr="oct_scene_path", name="Export scene path",
      description="Full path to directory where the exported scene is created",
      default="", maxlen=1024, subtype="FILE_PATH")
  
  StringProperty(attr="oct_scenename", name="Scene Name",
      description="Name of Octane scene (.ocs) to use",
      default="testscene", maxlen=1024)
      
  BoolProperty(attr="oct_cameramotionblur", name="Camera Motion Blur",
      description="Enable camera motion blur",
      default=False)

  BoolProperty(attr="oct_enablefadvan", name="Export next frame",
      description="Export next frame while rendering the actual, recomended to animation",
      default=True)


#  BoolProperty(attr="oct_enablebatchmode", name="Batch Mode",
#      description="Render in background (required for animations)",
#      default=False)
  
 
  IntProperty(attr="oct_batchmodespp", name="Batch mode max samples per pixel",
      description="Max number of samples per pixels in batch mode; 0 = ignore",
      default=128, min=0, soft_min=0)

  
  BoolProperty(attr="oct_waitrender", name="Wait for OCT",
      description="Wait for render to finish; load image into render results (required for animations)",
      default=False)
  
 
  # Add OCT Camera Lens Radius
  bpy.types.Camera.FloatProperty(attr="oct_lensradius", name="OCT DOF Lens Radius", 
      description="Octane camera lens radius for depth of field",
      default=0.015, min=0, max=10, soft_min=0, soft_max=10, precision=3)


  # Use some of the existing panels
  import properties_render
  properties_render.RENDER_PT_render.COMPAT_ENGINES.add('OCTANE_RENDER')
  properties_render.RENDER_PT_layers.COMPAT_ENGINES.add('OCTANE_RENDER')
  properties_render.RENDER_PT_dimensions.COMPAT_ENGINES.add('OCTANE_RENDER')
  properties_render.RENDER_PT_output.COMPAT_ENGINES.add('OCTANE_RENDER')
  properties_render.RENDER_PT_post_processing.COMPAT_ENGINES.add('OCTANE_RENDER')
  del properties_render
  
  import properties_material
  properties_material.MATERIAL_PT_context_material.COMPAT_ENGINES.add('OCTANE_RENDER')
  properties_material.MATERIAL_PT_diffuse.COMPAT_ENGINES.add('OCTANE_RENDER')
  properties_material.MATERIAL_PT_shading.COMPAT_ENGINES.add('OCTANE_RENDER')
  properties_material.MATERIAL_PT_transp.COMPAT_ENGINES.add('OCTANE_RENDER')
  properties_material.MATERIAL_PT_mirror.COMPAT_ENGINES.add('OCTANE_RENDER')
  del properties_material
  
  import properties_texture
  for member in dir(properties_texture):
    subclass = getattr(properties_texture, member)
    try:
      subclass.COMPAT_ENGINES.add('OCTANE_RENDER')
    except:
      pass
  del properties_texture
  
  import properties_data_camera
  for member in dir(properties_data_camera):
    subclass = getattr(properties_data_camera, member)
    try:
      subclass.COMPAT_ENGINES.add('OCTANE_RENDER')
    except:
      pass
  del properties_data_camera
  
  import properties_world
  properties_world.WORLD_PT_environment_lighting.COMPAT_ENGINES.add('OCTANE_RENDER')
  del properties_world

  import properties_data_lamp
  properties_data_lamp.DATA_PT_sunsky.COMPAT_ENGINES.add('OCTANE_RENDER')
  del properties_data_lamp

def oct_lensradius(self, context):
  if context.scene.render.engine == 'OCTANE_RENDER':  
    self.layout.split().column().prop(context.camera, "oct_lensradius", text="OCT Lens Radius")


class RenderButtonsPanel(bpy.types.Panel):
  bl_space_type = 'PROPERTIES'
  bl_region_type = 'WINDOW'
  bl_context = "render"
  
  def poll(self, context):
    return context.scene.render.engine in self.COMPAT_ENGINES

class RENDER_PT_octrender_options(RenderButtonsPanel):
  bl_label = "Octane Options"
  COMPAT_ENGINES = {'OCTANE_RENDER'}

  def draw(self, context):
    layout = self.layout
    scene = context.scene 

    split = layout.split()
    col = split.column()
    col.label(text="Full path to Octane's executable:")
    col.prop(scene, "oct_path", text="")
    col.label(text="Full path where the scene is exported:")
    col.prop(scene, "oct_scene_path", text="")
    col.prop(scene, "oct_scenename", text="Scene Name")
    split = layout.split()
    col = split.column()
    col.prop(scene, "oct_cameramotionblur")
    split = layout.split()
    col = split.column()
    col.prop(scene, "oct_enablefadvan")
    split = layout.split()
    col = split.column()
    col.prop(scene, "oct_waitrender")
    if scene.oct_waitrender:
      split = layout.split()
      col = split.column()
      col.prop(scene, "oct_batchmodespp", text="Samples")



class OctaneRender(bpy.types.RenderEngine):
  bl_idname = 'OCTANE_RENDER'
  bl_label = "Octane"
  
  def _octexport(self, scene,  basepath, basename, suns, envmap, frame):
    # Depends on Blender version used
    try:
        from Mathutils import Vector
    except ImportError:
        from mathutils import Vector
    from itertools import zip_longest

    ff = lambda f:format(f,'.6f').rstrip('0')

    print('OCTBP ===> Begin export')
    # Force an update to object matrices when rendering animations
#    scene.set_frame(scene.frame_current)
    scene.set_frame(frame)
    sdir = '{}/{}'.format(basepath,basename)
    export_objects=[]
    objname = '{}/{}.obj'.format(basepath,basename)
    basenameN=basename+'-%s' % (frame)
    basenameP=basename+'-%s' % (frame-1)
    objnameN = '{}/{}.obj'.format(basepath,basenameN)
    mtlnameP = '{}/{}.mtl'.format(basepath,basenameP)
    if os.path.isfile(objnameN):
      print('Previosly created, skip')
      os.remove(objname)
      os.rename(objnameN,objname)
      if os.path.isfile(mtlnameP):
        os.remove(mtlnameP)
    else:
      for obj in scene.objects:
        if not obj.restrict_render and obj.type in ['MESH', 'SURFACE', 'META', 'TEXT', 'CURVE'] and scene.layers[next((i for i in range(len(obj.layers)) if obj.layers[i]))]:
  #      if not obj.restrict_render and scene.layers[next((i for i in range(len(obj.layers)) if obj.layers[i]))]:
          print('OCTBP ===> Object: {}'.format(obj.name))
          export_objects.append(obj)
      write( objname, export_objects , scene,False,False,True,True,True,True,False,True,True,True,False,True,False,False,False)


  def render(self, scene):
    import subprocess 
    import time

    def error(error):
      self.update_stats("-=|ERROR|", error)
      # Hold the render results window while we display the error message
      while True:
        if self.test_break():
          raise Exception("Octane plug-in: "+error)
        time.sleep(0.1)
      
    print('OCTBP ===> plug-in starting...')
    self.update_stats("", "Octane: Exporting scene, please wait...")

    # Make sure we have lights in our scene
    suns = [f for f in scene.objects if f.type=="LAMP" and f.data and f.data.type=="SUN" and scene.layers[next((i for i in range(len(f.layers)) if f.layers[i]))] and (f.data.sky.use_sky or f.data.sky.use_atmosphere)]
    envmap = (scene.world and any(ts and hasattr(ts.texture,'image') for ts in scene.world.texture_slots))
#    emitters = any(m.users and m.emit for m in bpy.data.materials)
    
    basepath = scene.oct_scene_path

    if not os.path.exists(basepath):
      error("Cannot find scenes directory") 

    # Get scene name
    basename = scene.oct_scenename
    if os.path.isfile(scene.oct_scenename):
      basename = os.path.basename(scene.oct_scenename)
    basename = basename.split('/')[-1].split('\\')[-1]
    if not basename:
      error("Invalid scene filename") 
    basename = os.path.splitext(basename)[0]
    
    self._octexport(scene,  basepath, basename, suns, envmap, scene.frame_current)

    lensAperture = ''
    if scene.camera.data.oct_lensradius:
      lensAperture=' --cam-aperture {}'.format(scene.camera.data.oct_lensradius)
    fdist = (scene.camera.location-scene.camera.data.dof_object.location).magnitude if scene.camera.data.dof_object else scene.camera.data.dof_distance
    focalDepth = ''
    if fdist:
            focalDepth=' --cam-focaldepth {} '.format(fdist)
    
    samplesPerImage=''
    if scene.oct_waitrender:
            samplesPerImage='-s {}'.format(scene.oct_batchmodespp)

    proj_ocs = '{}/{}.ocs'.format(basepath,basename)
#    ocsFile=os.path.join(octaneProjectPath.val,extractBase(context_name[0])+'.ocs').replace("\\", "/")
#    objFile=objFileName
    objfile = '{}/{}.obj'.format(basepath,basename)
#    meshName=octaneProjectName.val
    meshnode=basename
#    outImage=pngFileName
    pngfile = '{}/{}.png'.format(basepath,basename)
    
    prefix=samplesPerImage+lensAperture+focalDepth
    frame=scene.frame_current
    camera_param = ''
    light_param=''
    axes = ['x','y','z']
    
    camera = scene.camera
    matrix = camera.matrix
 
    position = rotate90x(matrix_vect(matrix,[0.0,0.0,0.0,1.0]))
    target = rotate90x(matrix_vect(matrix,[0.0,0.0,-1.0,1.0]))
    up = rotate90x(matrix_vect(matrix,[0.0,1.0,0.0,0.0]))
    for i in range(3):
            camera_param += " --cam-pos-%s %f" % (axes[i],position[i])
    for i in range(3):
            camera_param += " --cam-target-%s %f" % (axes[i],target[i])
    for i in range(3):
            camera_param += " --cam-up-%s %f" % (axes[i],up[i])

    fov=camera.data.angle*180.0/3.1415926536
    camera_param += " --cam-fov %f" % fov
    if scene.oct_cameramotionblur:
#                if octaneInterpolateFrame==0:#If interpolate on next frame
#                        Blender.Set('curframe',frame+1)
#                elif frame>1:#Interpolate on previous frame
#                        Blender.Set('curframe',frame-1)
            scene.set_frame(scene.frame_current - 1)
            matrix = scene.camera.matrix
            
            position = rotate90x(matrix_vect(matrix,[0.0,0.0,0.0,1.0]))
            target = rotate90x(matrix_vect(matrix,[0.0,0.0,-1.0,1.0]))
            up = rotate90x(matrix_vect(matrix,[0.0,1.0,0.0,0.0]))
#                target = rotate90x(matrix * Vector([0, 0, -10]))

            for i in range(3):
                    camera_param += " --cam-motion-pos-%s %f" % (axes[i],position[i])
            for i in range(3):
                    camera_param += " --cam-motion-target-%s %f" % (axes[i],target[i])
            for i in range(3):
                    camera_param += " --cam-motion-up-%s %f" % (axes[i],up[i])
            scene.set_frame(scene.frame_current + 1)
    #Set the daylight option
#        if octaneExportSunDirection.val:
#                light = Blender.Object.Get(octaneActiveLightSource.val)
            #lightdata = light.getData()
#                matrix = Mathutils.Matrix(light.getInverseMatrix())
#                matrix = rotate90x(matrix)
            
#                axes = ['x','y','z']
#                for i in range(3):
#                        light_param += " --daylight-sundir-%s %f" % (axes[i],matrix[i][2])
#        octanePath=octaneRenderBinary.val
#        octanePath='\"'+octanePath+'\"'
    objPath=os.path.join(scene.oct_scene_path,objfile).replace("\\", "/")

    # Path where Octane is located
    if os.path.isdir(scene.oct_path):
      octanePath = basepath+'/octane'
    elif os.path.isfile(scene.oct_path):
      octanePath = scene.oct_path 
    else:
      error("Full path to Octane executable required")
    if not os.path.exists(octanePath):
      error("Cannot find Octane executable at specified path")


    projectFileAlreadyExists=False
    if os.path.exists(proj_ocs):
            projectFileAlreadyExists=True

    #If linux env add quotes in file names to allow spaces 
#        if sys.platform == "linux2" or sys.platform == "darwin":
#                objPath='\"'+objPath+'\"'
#                objfile='\"'+objfile+'\"'
#                proj_ocs='\"'+proj_ocs+'\"'
#                pngfile='\"'+pngfile+'\"'

    #Set the exit after rendering flag (for animation)
    #Also set the system flag to force blender to wait Octane has finished the rendering or to immediately
    #render the control to blender (for single frame rendering)
    exitFlag=''
    commandFlag=''
    if scene.oct_waitrender:
            exitFlag=' -e -q'
            #exitFlag=' -e'

#                if sys.platform == "win32":
#                        commandFlag=' /wait /normal \"\" '
#                elif sys.platform == "linux2" or sys.platform == "darwin":
#                        commandFlag = "(nice -n %d %s)"
#        else:
#                if sys.platform == "win32":
#                        commandFlag=' /b /normal \"\" '
                    
#                elif sys.platform == "linux2" or sys.platform == "darwin":
#                        commandFlag = "(nice -n %d %s)&"
            
    #Set the resolution option
    # Get resolution
    r = scene.render
    x = int(r.resolution_x*r.resolution_percentage*0.01)
    y = int(r.resolution_y*r.resolution_percentage*0.01)
    
    resolution=' --film-width %i --film-height %i' % (x,y)
            
    #Set the GPUs option
    useGPU=''
#        if octaneGPUSelector.val:
#                for val in octaneGPUUseList.val.split(' '):
#                        useGPU+=' -g '+val
    
                    
#        command =  octanePath+exitFlag+resolution+useGPU
    command =  exitFlag+resolution+useGPU
    # a falta de poner en el ui
    octaneReplaceProject=False
    #Use an existing octane ocs file
    if projectFileAlreadyExists and not octaneReplaceProject:
            if scene.oct_waitrender:
                    command+=' -m %s.obj -r %s %s --output-png %s %s %s %s' % (meshnode,objfile,prefix,pngfile,camera_param,light_param,proj_ocs)
            else:
                    command+=' -m %s.obj -r %s %s %s %s %s' % (meshnode,objfile,prefix,camera_param,light_param,proj_ocs)
            
    #Force creation of the octane ocs file if it doesn't exist      
    else:
            if scene.oct_waitrender:
                    command+=' -n %s -m %s.obj -l %s %s --output-png %s %s %s' % (proj_ocs,meshnode,objPath,prefix,pngfile,camera_param,light_param)
            else:
                    command+=' -n %s -m %s.obj -l %s %s %s %s' % (proj_ocs,meshnode,objPath,prefix,camera_param,light_param)
                    
            #Reset the replace project flag to avoid accidental replacement
            octaneReplaceProject=False
     
    #If a batch file should be created
#        if octaneWriteBatch.val:
#                file = open(filename,"a")
#                file.write(command + '\n')
#                file.close()

    #If no batch file should be created -> start octane immediately 
#        else:
#                if ostype == "win32":
#                        command='start ' + commandFlag + command                                                                                                                                                                  
#                        print command
#                        os.system(command)
                    
#                elif ostype == "linux2" or ostype == "darwin":
#                        if exitAfter:
#                                command=commandFlag%(0, command)
#                                print command
#                                subprocess.call(command,shell=True)
#                        else:
#                                command=commandFlag%(0, command)
#                                print command
#                                os.system(command)
                    
#        print('OCTBP ===> launch Octane: {} {}/{}.ocs'.format(exepath,basepath,basename))
    print('OCTBP ===> launch Octane: {} {} '.format(octanePath,command))
#        octproc = subprocess.Popen([octanePath,'{}'.format(command)], cwd=basepath, shell=False)
    octproc = subprocess.Popen([octanePath,'{}'.format(command)])
    
    if scene.oct_waitrender:
      # Wait for OCT , convert and load image
      self.update_stats("", "Octane: Batch Rendering frame# {} (see console for progress), please wait...".format(scene.frame_current))
      # Hold the render results window hostage until OCT returns...
      factual=scene.frame_current
      fnext=factual+1
      fadvan=scene.oct_enablefadvan
      if fnext > scene.frame_end:
        fadvan=False
      while True:
        if octproc.poll() != None:
          result = self.begin_result(0, 0, x, y)
          try:
            print('OCTBP ===> load image from file: {}'.format(pngfile))
            result.layers[0].load_from_file('{}'.format(pngfile))
          except:
            pass
          self.end_result(result)
          break
        if self.test_break():
          try:
            octproc.terminate()
          except:
            pass
          break
        time.sleep(1)
        if fadvan:
          basename+='-%s' % (fnext)
          print('preparing next frame {}'.format(basename))
          self._octexport(scene,  basepath, basename, suns, envmap, fnext)
#          print('restaura frame')
          scene.set_frame(factual)
          print('done')
          fadvan=False
    
    
        
def register():
  oct_properties()
  bpy.types.register(RENDER_PT_octrender_options)
  bpy.types.register(OctaneRender)
  bpy.types.DATA_PT_camera.append(oct_lensradius)
    
def unregister():
  bpy.types.unregister(RENDER_PT_octrender_options)
  bpy.types.unregister(OctaneRender)
  bpy.types.DATA_PT_camera.remove(oct_lensradius)

if __name__ == "__main__":
  register()